from typing import Any, List, Optional, Tuple, Type, Union, Callable
from types import ModuleType

from graia.saya.cube import Cube

from .schema import BaseSchema
from .context import channel_instance

class Channel:
    module: str
    _name: str = None
    _author: List[str]
    _description: Optional[str] = None

    _export: Any = None
    _py_module: ModuleType = None

    content: List[Cube]

    # TODO: _export reload for other modules

    def __init__(self, module: str) -> None:
        self.module = module
        self._author = []
        self.content = []

    def name(self, name: str):
        self._name = name
        return self

    def author(self, author: str):
        self._author.append(author)
        return self

    def description(self, description: str):
        self._description = description
        return self

    @staticmethod
    def current() -> "Channel":
        return channel_instance.get()

    def export(self, target):
        self._export = target
        return target

    def use(self, schema: BaseSchema):
        def use_wrapper(target: Union[Type, Callable, Any]):
            self.content.append(Cube(target, schema))
            return target
        return use_wrapper

    def cancel(self, target: Union[Type, Callable, Any]):
        self.content = [i for i in self.content if i.content is not target]



from typing import Any, Union

from graia.broadcast import Broadcast
from graia.saya.behaviour import Behaviour
from graia.saya.cube import Cube
from .schema import ListenerSchema

class BroadcastBehaviour(Behaviour):
    broadcast: Broadcast

    def __init__(self, broadcast: Broadcast) -> None:
        self.broadcast = broadcast

    def allocate(self, cube: Cube[Union[ListenerSchema,]]):
        if isinstance(cube.metaclass, ListenerSchema):
            listener = cube.metaclass.build_listener(cube.content)
            if not listener.namespace:
                listener.namespace = self.broadcast.getDefaultNamespace()
            self.broadcast.listeners.append(listener)
        else:
            return
        return True

    def uninstall(self, cube: Cube) -> Any:
        if isinstance(cube.metaclass, ListenerSchema):
            self.broadcast.removeListener(
                self.broadcast.getListener(cube.content)
            )
        else:
            return

        return True


from typing import Any, Generic, TypeVar
from .schema import BaseSchema
from dataclasses import dataclass

T = TypeVar("T", None, BaseSchema)

@dataclass(init=True)
class Cube(Generic[T]):
    content: Any
    metaclass: T


from dataclasses import dataclass, field
from typing import Optional, TYPE_CHECKING
from .context import channel_instance

if TYPE_CHECKING:
    from .channel import Channel

class BaseSchema:
    id: Optional[str] = None
    channel: "Channel" = field(default_factory=lambda: channel_instance.get(), init=False)


from abc import ABCMeta, abstractmethod
from typing import Any
from graia.saya.cube import Cube

class Behaviour(metaclass=ABCMeta):
    @abstractmethod
    def allocate(self, cube: Cube[Any]) -> Any:
        pass

    @abstractmethod
    def uninstall(self, cube: Cube[Any]) -> Any:
        pass


from dataclasses import dataclass
from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    from graia.saya.cube import Cube
    from .entity import Behaviour

@dataclass(init=True)
class RequireContext:
    module: str
    behaviours: List["Behaviour"]
    _index: int = 0

@dataclass(init=True)
class AllocationContext:
    cube: "Cube"


from dataclasses import dataclass, field
from typing import Callable, List, Type
from graia.broadcast import Listener, BaseEvent, BaseDispatcher
from graia.broadcast.entities.decorator import Decorator
from graia.broadcast.entities.namespace import Namespace
from graia.saya.schema import BaseSchema

@dataclass
class ListenerSchema(BaseSchema):
    listening_events: List[Type[BaseEvent]]
    namespace: Namespace = None
    inline_dispatchers: List[BaseDispatcher] = field(default_factory=list)
    headless_decorators: List[Decorator] = field(default_factory=list)
    priority: int = 16
    enable_internal_access: bool = False

    def build_listener(self, callable: Callable):
        return Listener(
            callable=callable,
            namespace=self.namespace,
            listening_events=self.listening_events,
            inline_dispatchers=self.inline_dispatchers,
            headless_decorators=self.headless_decorators,
            priority=self.priority,
            enable_internal_access=self.enable_internal_access
        )


from graia.broadcast.entities.event import BaseEvent
from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.interfaces.dispatcher import DispatcherInterface
from graia.saya.channel import Channel
from graia.saya.context import saya_instance
from typing import TYPE_CHECKING


class SayaModuleInstalled(BaseEvent):
    """不用返回 RemoveMe, 因为 Cube 在被 uninstall 时会被清理掉, 如果你用的是规范的 Saya Protocol 的话.
    """
    module: str
    channel: Channel

    class Dispatcher(BaseDispatcher):
        async def catch(interface: "DispatcherInterface"):
            from graia.saya import Saya

            if interface.annotation is Saya:
                return saya_instance.get()
            elif interface.annotation is Channel:
                return interface.event.channel

class SayaModuleUninstall(BaseEvent):
    """不用返回 RemoveMe, 因为 Cube 在被 uninstall 时会被清理掉, 如果你用的是规范的 Saya Protocol 的话.
    """
    module: str
    channel: Channel

    class Dispatcher(BaseDispatcher):
        async def catch(interface: "DispatcherInterface"):
            from graia.saya import Saya

            if interface.annotation is Saya:
                return saya_instance.get()
            elif interface.annotation is Channel:
                return interface.event.channel

class SayaModuleUninstalled(BaseEvent):
    """不用返回 RemoveMe, 因为 Cube 在被 uninstall 时会被清理掉, 如果你用的是规范的 Saya Protocol 的话.
    """
    module: str

    class Dispatcher(BaseDispatcher):
        async def catch(interface: "DispatcherInterface"):
            from graia.saya import Saya

            if interface.annotation is Saya:
                return saya_instance.get()
            elif interface.annotation is Channel:
                return interface.event.channel


import itertools
from typing import Any, Generator, List, TYPE_CHECKING

from graia.saya.cube import Cube
from .context import RequireContext, AllocationContext
from .entity import Behaviour

from graia.broadcast.exceptions import RequirementCrashed

if TYPE_CHECKING:
    from graia.saya import Saya

# TODO: Lifecycle for Behaviour

class BehaviourInterface:
    saya: "Saya"

    require_contents: List[RequireContext]
    allocation_contents: List[AllocationContext]

    def __init__(self, saya_instance: "Saya") -> None:
        self.saya = saya_instance
        self.require_contents = [RequireContext("graia.saya.__special__.global_behaviours", [])]
        self.allocation_contents = [AllocationContext(None)]

    @property
    def currentModule(self):
        return self.require_contents[-1].module
    
    @property
    def currentCube(self):
        return self.allocation_contents[-1].cube

    @property
    def _index(self):
        return self.require_contents[-1]._index

    def require_context(self, module: str, behaviours: List['Behaviour'] = None):
        self.require_contents.append(RequireContext(module, behaviours or []))
        return self

    def __enter__(self) -> "BehaviourInterface":
        return self
    
    def __exit__(self, _, exc: Exception, tb):
        self.require_contents.pop() # just simple.
        if tb is not None:
            raise exc.with_traceback(tb)
    
    def behaviour_generator(self):
        yield from self.require_contents[0].behaviours
        # Cube 没有 behaviours 设定, 哦, 连 always 都没有.
        yield from self.require_contents[-1].behaviours
    
    def allocate_cube(self, cube: Cube) -> Any: # = DispatcherInterface.lookup_param
        self.allocation_contents.append(AllocationContext(cube))
        
        try:
            start_offset = self._index + int(bool(self._index))
            for self.require_contents[-1]._index, behaviour in enumerate(
                itertools.islice(self.behaviour_generator(), start_offset, None, None),
                start=start_offset
            ):
                result = behaviour.allocate(cube)

                if result is None:
                    continue

                self.require_contents[-1]._index = 0
                return result
            else:
                raise RequirementCrashed(
                    f"the dispatching requirement crashed: {cube}"
                )
        finally:
            self.allocation_contents.pop()
    
    def uninstall_cube(self, cube: Cube) -> Any:
        self.allocation_contents.append(AllocationContext(cube))
        
        try:
            start_offset = self._index + int(bool(self._index))
            for self.require_contents[-1]._index, behaviour in enumerate(
                itertools.islice(self.behaviour_generator(), start_offset, None, None),
                start=start_offset
            ):
                result = behaviour.uninstall(cube)

                if result is None:
                    continue

                self.require_contents[-1]._index = 0
                return result
            else:
                raise RequirementCrashed(
                    f"the dispatching requirement crashed: {cube}"
                )
        finally:
            self.allocation_contents.pop()


import sys
import asyncio
import importlib
import sys
import copy
from typing import Any, Dict, List, NoReturn, Optional, Union
from contextlib import contextmanager

from graia.saya.channel import Channel

from graia.saya.schema import BaseSchema
from graia.saya.cube import Cube
from graia.saya.behaviour import Behaviour, BehaviourInterface
from graia.broadcast import Broadcast
from .context import saya_instance, channel_instance
from loguru import logger

class Saya:
    """Modular application for Graia Framework.

    > 名称取自作品 魔女之旅 中的角色 "沙耶(Saya)", 愿所有人的心中都有一位活泼可爱的炭之魔女.

    Saya 的架构分为: `Saya Controller`(控制器), `Module Channel`(模块容器), `Cube`(内容容器), `Behaviour`(行为).

     - `Saya Controller` 负责管理各个模块的

    Raises:
        TypeError: [description]

    Returns:
        [type]: [description]
    """
    behaviour_interface: BehaviourInterface
    behaviours: List[Behaviour]
    channels: Dict[str, Channel]
    broadcast: Optional[Broadcast] = None

    def __init__(self, broadcast: Broadcast = None) -> None:
        self.channels = {}
        self.behaviours = []
        self.behaviour_interface = BehaviourInterface(self)
        self.behaviour_interface.require_contents[0].behaviours = self.behaviours

        self.broadcast = broadcast
    
    @contextmanager
    def module_context(self):
        saya_token = saya_instance.set(self)
        yield
        saya_instance.reset(saya_token)

    @staticmethod
    def current() -> "Saya":
        return saya_instance.get()

    def require_resolve(self, module: str) -> Channel:
        channel = Channel(module)
        channel_token = channel_instance.set(channel)

        try:
            imported_module = importlib.import_module(module, module)
            channel._py_module = imported_module
            with self.behaviour_interface.require_context(module) as interface:
                for cube in channel.content:
                    try:
                        interface.allocate_cube(cube)
                    except:
                        logger.exception(f"an error occurred while loading the module's cube: {module}::{cube}")
                        raise
        finally:
            channel_instance.reset(channel_token)
        
        return channel

    def require(self, module: str) -> Union[Channel, Any]:
        logger.debug(f"require {module}")

        if module in self.channels:
            channel = self.channels[module]
            if channel._export:
                return channel._export
            return channel

        channel = self.require_resolve(module)
        self.channels[module] = channel

        if self.broadcast:
            token = saya_instance.set(self)
            self.broadcast.postEvent(SayaModuleInstalled(
                module=module,
                channel=channel,
            ))
            saya_instance.reset(token)

        logger.info(f"module loading finished: {module}")

        if channel._export:
            return channel._export

        return channel
    
    def install_behaviours(self, *behaviours: Behaviour):
        self.behaviours.extend(behaviours)
    
    def uninstall_channel(self, channel: Channel):
        if channel not in self.channels.values():
            raise TypeError("assert an existed channel")
        
        if channel.module == "__main__":
            raise ValueError("main channel cannot uninstall")

        # TODO: builtin signal(async or sync)
        if self.broadcast:
            token = saya_instance.set(self)
            self.broadcast.postEvent(SayaModuleUninstall(
                module=channel.module,
                channel=channel,
            ))
            saya_instance.reset(token)

        with self.behaviour_interface.require_context(channel.module) as interface:
            for cube in channel.content:
                try:
                    interface.uninstall_cube(cube)
                except:
                    logger.exception(f"an error occurred while loading the module's cube: {channel.module}::{cube}")
                    raise
        
        del self.channels[channel.module]
        channel._py_module = None

        if sys.modules.get(channel.module):
            del sys.modules[channel.module]

        if self.broadcast:
            token = saya_instance.set(self)
            self.broadcast.postEvent(SayaModuleUninstalled(
                module=channel.module,
            ))
            saya_instance.reset(token)
    
    def reload_channel(self, channel: Channel = None, module: str = None) -> None:
        if not channel and not module:
            raise TypeError("require a Channel or string.")

        if not channel and module:
            channel = self.channels.get(module)
            if not channel:
                raise ValueError("cannot found the module.")

        self.uninstall_channel(channel)
        new_channel = self.require_resolve(channel.module)

        channel._name = new_channel._name
        channel._author = new_channel._author
        channel._description = new_channel._description
        channel._export = new_channel._export
        channel._py_module = new_channel._py_module
        channel.content = new_channel.content
    
    def create_main_channel(self) -> Channel:
        may_current = self.channels.get("__main__")
        if may_current:
            return may_current

        main_channel = Channel("__main__")
        self.channels["__main__"] = main_channel
        
        if self.broadcast:
            token = saya_instance.set(self)
            self.broadcast.postEvent(SayaModuleInstalled(
                module="__main__",
                channel=main_channel,
            ))
            saya_instance.reset(token)
        
        return main_channel

from .event import SayaModuleInstalled, SayaModuleUninstall, SayaModuleUninstalled